@@ -2,17 +2,20 @@ package Plack::Request;
 use strict;
 use warnings;
 use 5.008_001;
-our $VERSION = '1.0039';
+our $VERSION = '1.0041';
 
 use HTTP::Headers::Fast;
 use Carp ();
 use Hash::MultiValue;
-use HTTP::Body;
 
 use Plack::Request::Upload;
 use Stream::Buffered;
 use URI;
 use URI::Escape ();
+use Cookie::Baker ();
+
+use HTTP::Entity::Parser;
+use WWW::Form::UrlEncoded qw/parse_urlencoded_arrayref/;
 
 sub new {
     my($class, $env) = @_;
@@ -59,41 +62,7 @@ sub cookies {
     }
 
     $self->env->{'plack.cookie.string'} = $self->env->{HTTP_COOKIE};
-
-    my %results;
-    my @pairs = grep m/=/, split "[;,] ?", $self->env->{'plack.cookie.string'};
-    for my $pair ( @pairs ) {
-        # trim leading trailing whitespace
-        $pair =~ s/^\s+//; $pair =~ s/\s+$//;
-
-        my ($key, $value) = map URI::Escape::uri_unescape($_), split( "=", $pair, 2 );
-
-        # Take the first one like CGI.pm or rack do
-        $results{$key} = $value unless exists $results{$key};
-    }
-
-    $self->env->{'plack.cookie.parsed'} = \%results;
-}
-
-sub query_parameters {
-    my $self = shift;
-    $self->env->{'plack.request.query'} ||= $self->_parse_query;
-}
-
-sub _parse_query {
-    my $self = shift;
-
-    my @query;
-    my $query_string = $self->env->{QUERY_STRING};
-    if (defined $query_string) {
-        $query_string =~ s/\A[&;]+//;
-        @query =
-            map { s/\+/ /g; URI::Escape::uri_unescape($_) }
-            map { /=/ ? split(/=/, $_, 2) : ($_ => '')}
-            split(/[&;]+/, $query_string);
-    }
-
-    Hash::MultiValue->new(@query);
+    $self->env->{'plack.cookie.parsed'} = Cookie::Baker::crush_cookie($self->env->{'plack.cookie.string'});
 }
 
 sub content {
@@ -137,14 +106,27 @@ sub header           { shift->headers->header(@_) }
 sub referer          { shift->headers->referer(@_) }
 sub user_agent       { shift->headers->user_agent(@_) }
 
-sub body_parameters {
+sub _body_parameters {
     my $self = shift;
-
-    unless ($self->env->{'plack.request.body'}) {
+    unless ($self->env->{'plack.request.body_parameters'}) {
         $self->_parse_request_body;
     }
+    return $self->env->{'plack.request.body_parameters'};
+}
 
-    return $self->env->{'plack.request.body'};
+sub _query_parameters {
+    my $self = shift;
+    $self->env->{'plack.request.query_parameters'} ||= parse_urlencoded_arrayref($self->env->{'QUERY_STRING'});
+}
+
+sub query_parameters {
+    my $self = shift;
+    $self->env->{'plack.request.query'} ||= Hash::MultiValue->new(@{$self->_query_parameters});
+}
+
+sub body_parameters {
+    my $self = shift;
+    $self->env->{'plack.request.body'} ||= Hash::MultiValue->new(@{$self->_body_parameters});
 }
 
 # contains body + query
@@ -152,9 +134,10 @@ sub parameters {
     my $self = shift;
 
     $self->env->{'plack.request.merged'} ||= do {
-        my $query = $self->query_parameters;
-        my $body  = $self->body_parameters;
-        Hash::MultiValue->new($query->flatten, $body->flatten);
+        Hash::MultiValue->new(
+            @{$self->_query_parameters},
+            @{$self->_body_parameters}
+        );
     };
 }
 
@@ -237,78 +220,56 @@ sub new_response {
     Plack::Response->new(@_);
 }
 
-sub _parse_request_body {
+sub request_body_parser {
+    my $self = shift;
+    $self->{request_body_parser} ||= $self->_build_body_parser;
+}
+
+sub _build_body_parser {
     my $self = shift;
 
-    my $ct = $self->env->{CONTENT_TYPE};
-    my $cl = $self->env->{CONTENT_LENGTH};
-    if (!$ct && !$cl) {
-        # No Content-Type nor Content-Length -> GET/HEAD
-        $self->env->{'plack.request.body'}   = Hash::MultiValue->new;
-        $self->env->{'plack.request.upload'} = Hash::MultiValue->new;
+    my $len = $self->_buffer_length_for($self->env);
+
+    my $parser = HTTP::Entity::Parser->new(buffer_length => $len);
+    $parser->register('application/x-www-form-urlencoded', 'HTTP::Entity::Parser::UrlEncoded');
+    $parser->register('multipart/form-data', 'HTTP::Entity::Parser::MultiPart');
+
+    $parser;
+}
+
+sub _buffer_length_for {
+    my($self, $env) = @_;
+
+    return $ENV{PLACK_BUFFER_LENGTH} if defined $ENV{PLACK_BUFFER_LENGTH};
+
+    if ($env->{'psgix.input.buffered'}) {
+        return 1024 * 1024; # 1MB for buffered
+    } else {
+        return 1024 * 64; # 64K for unbuffered
+    }
+}
+
+sub _parse_request_body {
+    my $self = shift;
+    if ( !$self->env->{CONTENT_TYPE} ) {
+        $self->env->{'plack.request.body_parameters'} = [];
+        $self->env->{'plack.request.upload'} = Hash::MultiValue->new();
         return;
     }
 
-    my $body = HTTP::Body->new($ct, $cl);
+    my ($params,$uploads) = $self->request_body_parser->parse($self->env);
+    $self->env->{'plack.request.body_parameters'} = $params;
 
-    # HTTP::Body will create temporary files in case there was an
-    # upload.  Those temporary files can be cleaned up by telling
-    # HTTP::Body to do so. It will run the cleanup when the request
-    # env is destroyed. That the object will not go out of scope by
-    # the end of this sub we will store a reference here.
-    $self->env->{'plack.request.http.body'} = $body;
-    $body->cleanup(1);
-
-    my $input = $self->input;
-
-    my $buffer;
-    if ($self->env->{'psgix.input.buffered'}) {
-        # Just in case if input is read by middleware/apps beforehand
-        $input->seek(0, 0);
-    } else {
-        $buffer = Stream::Buffered->new($cl);
+    my $upload_hash = Hash::MultiValue->new();
+    while ( my ($k,$v) = splice @$uploads, 0, 2 ) {
+        my %copy = %$v;
+        $copy{headers} = HTTP::Headers::Fast->new(@{$v->{headers}});
+        $upload_hash->add($k, Plack::Request::Upload->new(%copy));
     }
-
-    my $spin = 0;
-    while ($cl) {
-        $input->read(my $chunk, $cl < 8192 ? $cl : 8192);
-        my $read = length $chunk;
-        $cl -= $read;
-        $body->add($chunk);
-        $buffer->print($chunk) if $buffer;
-
-        if ($read == 0 && $spin++ > 2000) {
-            Carp::croak "Bad Content-Length: maybe client disconnect? ($cl bytes remaining)";
-        }
-    }
-
-    if ($buffer) {
-        $self->env->{'psgix.input.buffered'} = 1;
-        $self->env->{'psgi.input'} = $buffer->rewind;
-    } else {
-        $input->seek(0, 0);
-    }
-
-    $self->env->{'plack.request.body'}   = Hash::MultiValue->from_mixed($body->param);
-
-    my @uploads = Hash::MultiValue->from_mixed($body->upload)->flatten;
-    my @obj;
-    while (my($k, $v) = splice @uploads, 0, 2) {
-        push @obj, $k, $self->_make_upload($v);
-    }
-
-    $self->env->{'plack.request.upload'} = Hash::MultiValue->new(@obj);
-
+    $self->env->{'plack.request.upload'} = $upload_hash;
     1;
 }
 
-sub _make_upload {
-    my($self, $upload) = @_;
-    my %copy = %$upload;
-    $copy{headers} = HTTP::Headers::Fast->new(%{$upload->{headers}});
-    Plack::Request::Upload->new(%copy);
-}
-
 1;
 __END__
 
@@ -348,9 +309,16 @@ certainly possible but not recommended: it's like doing so with
 mod_perl's Apache::Request: yet too low level.
 
 If you're writing a web application, not a framework, then you're
-encouraged to use one of the web application frameworks that support PSGI (L<http://plackperl.org/#frameworks>),
-or see modules like L<HTTP::Engine> to provide higher level
-Request and Response API on top of PSGI.
+encouraged to use one of the web application frameworks that support
+PSGI (L<http://plackperl.org/#frameworks>), or see modules like
+L<HTTP::Engine> to provide higher level Request and Response API on
+top of PSGI.
+
+If you're looking for an easy-to-use API to convert existing CGI
+applications to run on PSGI, consider using L<CGI::PSGI> or
+L<CGI::Emulate::PSGI> as well. L<CGI::Emulate::PSGI> documentation has
+a good summary of using them to convert existing CGI scripts to adapt
+to PSGI.
 
 =head1 METHODS
 
@@ -458,7 +426,7 @@ strings that are sent by clients and are URI decoded.
 If there are multiple cookies with the same name in the request, this
 method will ignore the duplicates and return only the first value. If
 that causes issues for you, you may have to use modules like
-CGI::Simple::Cookie to parse C<< $request->header('Cookies') >> by
+CGI::Simple::Cookie to parse C<< $request->header('Cookie') >> by
 yourself.
 
 =item query_parameters
